from otree_tools.models import Exit, FocusEvent
import tempfile
from django.urls import reverse
from django.template.loader import render_to_string
from channels import Group as ChannelGroup
import json
from otree_tools import cp
tracker_correspondence = {'time': Exit,
                          'focus_per_page': FocusEvent,
                          'focus_raw': FocusEvent}


class FileMaker:
    button_block = 'otree_tools/includes/initial_export_button.html'
    model = None

    def __init__(self, channel_to_response, tracker_type):
        self.model = tracker_correspondence[tracker_type]
        self.channel_to_response = channel_to_response
        self.tracker_type = tracker_type
        self.template_name = f'otree_tools/trackers_export/{tracker_type}_data.csv'
        self.channel = ChannelGroup(self.channel_to_response)

    def get_data(self):
        data_method = getattr(self.model.export, self.tracker_type)
        c = {
            'events': data_method(),
        }
        fp = tempfile.NamedTemporaryFile('w', prefix=f'{self.tracker_type}_',
                                         suffix='.csv',
                                         encoding='utf-8',
                                         delete=False)
        fp.write(render_to_string(self.template_name, c))
        fp.flush()
        link_to_send = reverse('export_tracker_data', kwargs={'temp_file_name': fp.name,
                                                              'tracker_type': self.tracker_type})

        self.channel.send({'text': json.dumps({'url': link_to_send,
                                               'button': render_to_string(self.button_block)})
                           })



import vanilla
import csv
import os
import json
from datetime import datetime

from django.db.models import ExpressionWrapper, F, DurationField
from django.views.generic import ListView, TemplateView, View
from django.http import HttpResponse
from django.core.urlresolvers import reverse

from otree.models import Session, Participant
from otree.views.export import get_export_response

from .models import Exit, FocusEvent, allowed_export_tracker_requests
from .export import export_wide
from . import __version__ as otree_tools_version


# TIME STAMPS VIEWS FOR TRACKING_TIME and TRACKING_FOCUS
class EnterExitMixin:
    def get_queryset(self):
        # TODO: group by page name?
        tot_exits = Exit.objects.filter(
            enter__isnull=False,
        ).annotate(diff=ExpressionWrapper(F('timestamp') - F('enter__timestamp'),
                                          output_field=DurationField()))
        for i in tot_exits:
            if i.diff is None:
                i.diff = i.timestamp - i.enter.timestamp
        return tot_exits


class PaginatedListView(ListView):
    navbar_active_tag = None
    export_link_name = None

    def get_context_data(self, **kwargs):
        c = super().get_context_data(**kwargs)
        c['nbar'] = self.navbar_active_tag
        curpage_num = c['page_obj'].number
        paginator = c['paginator']
        epsilon = 3
        c['allowed_range'] = range(max(1, curpage_num - epsilon), min(curpage_num + epsilon, paginator.num_pages) + 1)
        if self.export_link_name:
            c['export_link'] = reverse(self.export_link_name)
        return c


class EnterExitEventList(EnterExitMixin, PaginatedListView):
    template_name = 'otree_tools/trackers_export/all_timespent_list.html'
    url_name = 'time_spent_timestamps'
    url_pattern = r'^time_spent_per_page/$'
    display_name = 'Time spent per page [otree-tools]'
    context_object_name = 'timestamps'
    paginate_by = 50
    navbar_active_tag = 'time'


class TempFileCSVExport(View):
    """Receives a temp file name prepared in consumers, and returns it to a user.
    That is mostly for dealing with large files that can provoke server timeout errors."""
    url_name = 'export_tracker_data'
    url_pattern = fr'^export_tracker_data/(?P<tracker_type>{allowed_export_tracker_requests})/(?P<temp_file_name>.*)/$'
    content_type = 'text/csv'

    def get(self, request, *args, **kwargs):
        file_path = kwargs.get('temp_file_name')
        tracker_type = kwargs.get('tracker_type')
        curtime = datetime.now().strftime("%d_%m_%Y__%H_%M_%S")
        filename = f'{tracker_type}_{curtime}.csv'

        fsock = open(file_path, "r")
        response = HttpResponse(fsock, content_type=self.content_type)
        response['Content-Disposition'] = f'attachment; filename="{filename}"'
        fsock.close()
        try:
            os.remove(file_path)
        except FileNotFoundError:
            pass
            # TODO logger.warning('temp file not found')
        return response


########### BLOCK: FOCUS EVENTS EXPORT BLOCK ##############################################################

class FocusPerPageReport(PaginatedListView):
    template_name = 'otree_tools/trackers_export/focus_per_page_report.html'
    url_name = 'focus_per_page_report'
    url_pattern = r'^focus_per_page_report/$'
    display_name = 'Focused/unfocused [otree-tools]'
    context_object_name = 'focusevents'
    paginate_by = 50
    navbar_active_tag = 'focus'


    def get_queryset(self):
        return FocusEvent.objects.get_per_page_report()


class FocusEventList(PaginatedListView):
    template_name = 'otree_tools/trackers_export/focus_event_list.html'
    url_name = 'focus_events'
    url_pattern = r'^focus_events/$'
    display_name = 'Focus/unfocus events [otree-tools]'
    context_object_name = 'focusevents'
    queryset = FocusEvent.objects.all()
    paginate_by = 50
    navbar_active_tag = 'focus'


############ END OF: FOCUS EVENTS EXPORT BLOCK #############################################################


# END OF TIME STAMPS BLOCK


# Welcoming page
class HomeView(TemplateView):
    template_name = 'otree_tools/home.html'
    url_name = 'otree_tools_home'
    url_pattern = r'^otree_tools/$'
    display_name = 'otree-tools Home page'
    navbar_active_tag = 'home'

    def get_context_data(self, **kwargs):
        c = super().get_context_data(**kwargs)
        c['nbar'] = self.navbar_active_tag
        c['version'] = otree_tools_version
        return c


########### BLOCK: EXPORTING PARTICIPANT.VARS MODULE ##############################################################


class UniversalEncoder(json.JSONEncoder):
    """A 'universal encoder for exporting participant.vars."""

    def default(self, obj):
        if isinstance(obj, (set, range)):
            return list(obj)
        try:
            return json.JSONEncoder.default(self, obj)
        except TypeError:
            try:
                return str(obj)
            # the last resort - if built-in __str__ of the object causes an error we fail silently
            except:
                return ''


class PVarsMixin(object):
    """Mixin to use for exporting participant.vars both for text and html."""

    def get_session(self):
        return Session.objects.get(pk=self.kwargs.get('pk'))

    def get_queryset(self):
        headers_to_add = ['session.code', 'player.code']
        session = self.get_session()
        data = Participant.objects.filter(session=session).values('vars', 'code', 'session__code')
        # todo: sort the headers
        keys = sorted(list(set().union(*(d['vars'].keys() for d in data))), key=lambda s: s.lower())
        self.heads = headers_to_add + keys
        q = list()
        for i in data:
            dict_to_row = json.loads(json.dumps(i['vars'], cls=UniversalEncoder))
            row = [i['session__code'], i['code']] + [dict_to_row.get(key, '') for key in keys]
            q.append(row)

        return q


class ListPVarsView(PVarsMixin, PaginatedListView):
    """Returns  all participant vars in the database - per session."""
    template_name = 'otree_tools/pvars_list.html'
    url_name = 'pvars_list'
    url_pattern = r'^session/(?P<pk>[a-zA-Z0-9_-]+)/pvars/$'
    model = Participant
    context_object_name = 'pvars'
    paginate_by = 20
    navbar_active_tag = 'export'
    export_link_name = None

    def get_context_data(self, **kwargs):
        c = super().get_context_data(**kwargs)
        session = self.get_session()
        c['session'] = session
        c['heads'] = self.heads
        c['export_link'] = reverse('pvars_csv_export', kwargs={'pk': session.pk})
        return c


class PVarsCSVExport(PVarsMixin, TemplateView):
    """Returns the textual file   all participant vars in the database - per session."""
    url_name = 'pvars_csv_export'
    url_pattern = r'^session/(?P<pk>[a-zA-Z0-9_-]+)/pvars_csv_export/$'
    response_class = HttpResponse
    content_type = 'text/csv'

    def get(self, request, *args, **kwargs):
        response = HttpResponse(content_type='text/csv')
        session_code = self.get_session().code
        filename = f'{session_code}_participant_vars.csv'
        response['Content-Disposition'] = 'attachment; filename="{}"'.format(filename)

        pvars = self.get_queryset()
        writer = csv.writer(response)
        writer.writerow(self.heads)

        for p in pvars:
            writer.writerow(p)

        return response


############ END OF: EXPORTING PARTICIPANT.VARS MODULE #############################################################


########### BLOCK: SPECIFIC SESSION EXPORT ##############################################################


class AllSessionsList(PaginatedListView):
    template_name = 'otree_tools/all_session_list.html'
    url_name = 'individual_sessions_export'
    url_pattern = r'^individual_sessions_export/$'
    display_name = 'Exporting data from individual sessions'
    navbar_active_tag = 'export'
    queryset = Session.objects.all()
    context_object_name = 'sessions'
    paginate_by = 10


class SpecificSessionDataView(vanilla.TemplateView):
    url_pattern = r'^session_data/(?P<session_code>.*)/(?P<filetype>.*)/$'
    url_name = 'session_data'

    def get(self, request, *args, **kwargs):
        session_code = kwargs['session_code']
        response, file_extension = get_export_response(
            request, session_code)
        export_wide(response, file_extension, session_code=session_code)
        return response

############ END OF: SPECIFIC SESSION EXPORT #############################################################



from otree.api import Currency as c, currency_range
from . import pages
from ._builtin import Bot
from .models import Constants


class PlayerBot(Bot):

    def play_round(self):
        yield (pages.MyPage)
        yield (pages.Results)



from django.db import models
from otree.models import Participant
from django.db import transaction
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey
from otree_tools import cp
from django.db.models import F, ExpressionWrapper, DurationField, Sum, Min, Case, When

from datetime import timedelta

EXITTYPES = [(0, 'form submitted'), (1, 'page unloaded'), (2, 'client disconnected')]
FOCUS_ENTER_EVENT_TYPES = [(0, 'Page shown'), (3, 'Visibility: on'), (4, 'Focus: on'), ]
FOCUS_EXIT_EVENT_TYPES = [(1, 'Visibility: off'), (2, 'Focus: off'), (5, 'Form submitted'), ]
FOCUS_EVENT_TYPES = FOCUS_ENTER_EVENT_TYPES + FOCUS_EXIT_EVENT_TYPES
focus_exit_codes = [i for i, j in FOCUS_EXIT_EVENT_TYPES]
focus_enter_codes = [i for i, j in FOCUS_ENTER_EVENT_TYPES]
WAITFORIMAGES_CHOICES = [(False, 'Before images are loaded'), (True, 'After images are loaded')]
allowed_export_tracker_requests = r'(time|focus_per_page|focus_raw)'
"""
There are 3 different scenarios how a client may exit the page.
1. He can submit the form by clicking next (or in oTree any other button because the entire page
is wrapped into form tags
2. He can refresh the page (clicking F5 or command+R)
3. The browser can be closed (by accident or intentionally)

The Db structure is the following:
Enter timestamp is linked to participant, and it also contains info on page. All timestamps correspond on time
when a participant connects to a channel on page load.

since there are several exit timestamp events that can be linked to one single enter event, that means that they
should be connected to an enter timestamp via ForeignKey.
How the connection takes place:
The Enter event has  a boolean attribute Closed, initially set to false.
When the client disconnects it looks for all not closed enter events for this participant-page pair, and closes them
As a safeguard, the same closing happens everytime a client connects to the page, and before a new open Enter event is
created.

The aggregation for stats purposes is done the following way:
* we choose all enter events for the specific participant and page
* collect the earliest exit event for each enter event (via subquery and aggregation)
* annotate the query via F and ExpressionWrapper to calculate the difference between this earlist exit and its parent
Enter event
...
PROFIT!!
"""


class GeneralEvent(models.Model):
    class Meta:
        get_latest_by = 'timestamp'
        abstract = True

    page_name = models.CharField(max_length=1000)
    participant = models.ForeignKey(to=Participant, related_name="%(app_label)s_%(class)s_events")
    app_name = models.CharField(max_length=1000)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    player_id = models.PositiveIntegerField()
    player = GenericForeignKey('content_type', 'player_id')
    timestamp = models.DateTimeField()
    round_number = models.PositiveIntegerField()


class Enter(GeneralEvent):
    wait_for_images = models.BooleanField(choices=WAITFORIMAGES_CHOICES, default=True)

    def __str__(self):
        return f'id: {self.pk}, Enter: {self.page_name}; time: {self.timestamp}'


class ExitExportManager(models.Manager):
    """Manager for correctly exporting data to templates and csv from time trackers."""

    def time(self):
        csv_data = super().get_queryset().filter(
            enter__isnull=False,
        ).annotate(diff=ExpressionWrapper(F('timestamp') - F('enter__timestamp'),
                                          output_field=DurationField()))

        return csv_data


class Exit(GeneralEvent):
    objects = models.Manager()
    export = ExitExportManager()
    enter = models.OneToOneField(to='Enter', related_name='exit', null=True)
    timestamp = models.DateTimeField()
    exit_type = models.IntegerField(choices=EXITTYPES)

    def __str__(self):
        return f'id: {self.pk}, Exit: {self.page_name}; time: {self.timestamp}; Type: {self.get_exit_type_display()} '


class FocusExportManager(models.Manager):
    def focus_per_page(self):
        return self.model.objects.get_per_page_report()

    def focus_raw(self):
        return super().get_queryset()


class FocusManager(models.Manager):
    def _get_tracked_time(self, player_id, participant_id, page_name, focus_type):
        tot_exits = super().get_queryset().values('participant',
                                                  'page_name',
                                                  'player_id'). \
            filter(player_id=player_id,
                   participant_id=participant_id,
                   page_name=page_name,
                   entry__isnull=False,
                   ). \
            aggregate(
            diff=Sum(Case(
                When(event_num_type__in=focus_type,
                     then=ExpressionWrapper(F('timestamp') - F('entry__timestamp'),
                                            output_field=DurationField())),
                output_field=DurationField(),
            ))
        )

        return tot_exits.get('diff') or timedelta()

    def get_focused_time_per_page(self, player_id, participant_id, page_name):
        return self._get_tracked_time(player_id, participant_id, page_name, focus_exit_codes)

    def get_unfocused_time_per_page(self, player_id, participant_id, page_name):
        return self._get_tracked_time(player_id, participant_id, page_name,
                                      focus_enter_codes)

    def get_per_page_report(self):
        q = super().get_queryset().values('participant',
                                          'app_name',
                                          'page_name',
                                          'player_id',
                                          'participant__code',
                                          'participant__session__code',
                                          'round_number'). \
            filter(entry__isnull=False). \
            annotate(
            focused_time=Sum(Case(
                When(event_num_type__in=focus_exit_codes,
                     then=ExpressionWrapper(F('timestamp') - F('entry__timestamp'),
                                            output_field=DurationField())),
                output_field=DurationField(),
            )),
            unfocused_time=Sum(Case(
                When(event_num_type__in=focus_enter_codes,
                     then=ExpressionWrapper(F('timestamp') - F('entry__timestamp'),
                                            output_field=DurationField())),
                output_field=DurationField(),
            )),

        )
        return q


class FocusEvent(GeneralEvent):
    objects = FocusManager()
    export = FocusExportManager()
    event_desc_type = models.CharField(max_length=1000)
    event_num_type = models.IntegerField(choices=FOCUS_EVENT_TYPES)
    # we track focus events that have corresponding closuring focus event to
    # correctly calculate focus on/focus off time
    entry = models.OneToOneField(to='FocusEvent',
                                 related_name='closure',
                                 null=True,
                                 )

    def __str__(self):
        return f'id: {self.pk}, event_desc_type: {self.event_desc_type}; time: {self.timestamp};'





from otree.models import Participant, Session
from otree.export import get_field_names_for_csv, sanitize_for_csv, get_rows_for_wide_csv_round, get_models_module, \
    get_payoff_plus_participation_fee, _export_xlsx,_export_csv
from otree.session import SessionConfig
import collections
from collections import OrderedDict
from django.conf import settings
from django.db.models import Max



def get_rows_for_wide_csv(session_code):
    cursession=Session.objects.get(code=session_code)
    sessions = [cursession]
    session_cache = {row.id: row for row in sessions}
    session_config_fields = set()
    for session in sessions:
        for field_name in SessionConfig(session.config).editable_fields():
            session_config_fields.add(field_name)
    session_config_fields = list(session_config_fields)

    participants = Participant.objects.filter(session=cursession).order_by('id').values()
    if not participants:
        # 1 empty row
        return [[]]

    session_fields = get_field_names_for_csv(Session)
    participant_fields = get_field_names_for_csv(Participant)
    participant_fields.append('payoff_plus_participation_fee')
    header_row = ['participant.{}'.format(fname) for fname in participant_fields]
    header_row += ['session.{}'.format(fname)
                   for fname in session_fields]
    header_row += ['session.config.{}'.format(fname)
                   for fname in session_config_fields]
    rows = [header_row]
    for participant in participants:
        session = cursession #session_cache[participant['session_id']]
        participant['payoff_plus_participation_fee'] = get_payoff_plus_participation_fee(session, participant)
        row = [sanitize_for_csv(participant[fname]) for fname in participant_fields]

        row += [sanitize_for_csv(getattr(session, fname)) for fname in session_fields]
        row += [sanitize_for_csv(session.config.get(fname)) for fname in session_config_fields]
        rows.append(row)

    # heuristic to get the most relevant order of apps
    app_sequences = collections.Counter()
    for session in sessions:
        # we loaded the config earlier
        app_sequence = session.config['app_sequence']
        app_sequences[tuple(app_sequence)] += session.num_participants
    most_common_app_sequence = app_sequences.most_common(1)[0][0]

    apps_not_in_popular_sequence = [
        app for app in settings.INSTALLED_OTREE_APPS
        if app not in most_common_app_sequence]

    order_of_apps = list(most_common_app_sequence) + apps_not_in_popular_sequence

    rounds_per_app = OrderedDict()
    for app_name in order_of_apps:
        models_module = get_models_module(app_name)
        agg_dict = models_module.Subsession.objects.all().aggregate(Max('round_number'))
        highest_round_number = agg_dict['round_number__max']

        if highest_round_number is not None:
            rounds_per_app[app_name] = highest_round_number
    for app_name in rounds_per_app:
        for round_number in range(1, rounds_per_app[app_name] + 1):
            new_rows = get_rows_for_wide_csv_round(app_name, round_number, sessions)
            for i in range(len(rows)):
                try:
                    rows[i].extend(new_rows[i])
                except:
                    ...
    return rows

def export_wide(fp, file_extension='csv', session_code=None):
    rows = get_rows_for_wide_csv(session_code=session_code)
    if file_extension == 'xlsx':
        _export_xlsx(fp, rows)
    else:
        _export_csv(fp, rows)



# -*- coding: utf-8 -*-

from django.core.exceptions import ValidationError
from django.db.models import TextField

from .compat import add_meta_class, get_val_from_obj
from .fields import RadioGridFormField


class RadioGridField(TextField):
    def __init__(self, *args, **kwargs):
        self.rows = kwargs.pop('rows')
        self.values = kwargs.pop('values')
        self.require_all_fields = kwargs.pop('require_all_fields', True)
        self.verbose_name = kwargs.get('verbose_name', '')
        super(RadioGridField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(RadioGridField, self).deconstruct()
        kwargs['rows'] = self.rows
        kwargs['values'] = self.values
        kwargs['require_all_fields'] = self.require_all_fields
        return name, path, args, kwargs

    def from_db_value(self, value, expression, connection, context):
        return self.to_python(value)

    def to_python(self, value):
        return value if isinstance(value, list) else value.split(',')

    def get_prep_value(self, value):
        return '' if value is None else ','.join(value)

    def formfield(self, **kwargs):
        defaults = {
            'rows': self.rows,
            'values': self.values,
            'require_all_fields': self.require_all_fields,
            'label': self.verbose_name,
            'required': self.require_all_fields,
        }
        defaults.update(kwargs)
        return RadioGridFormField(**defaults)

    def value_to_string(self, obj):
        return self.get_prep_value(get_val_from_obj(self, obj))

    def validate(self, value, model_instance):
        allowed_values = [str(key) for key, _ in self.values]
        if not self.require_all_fields:
            allowed_values += ['']

        for v in value:
            if str(v) not in allowed_values:
                raise ValidationError(self.error_messages['invalid_choice'] % {"value": value})

    def contribute_to_class(self, cls, name, **kwargs):
        super(RadioGridField, self).contribute_to_class(cls, name, **kwargs)

        def get_list(obj):
            values = dict(self.values)
            display = []
            for value in getattr(obj, name):
                item = values.get(value, None)
                if item is None:
                    try:
                        item = values.get(int(value), value)
                    except (ValueError, TypeError):
                        item = value
                display.append(item)
            return display

        def get_display(obj):
            return ', '.join(get_list(obj))

        setattr(cls, 'get_%s_list' % self.name, get_list)
        setattr(cls, 'get_%s_display' % self.name, get_display)


RadioGridField = add_meta_class(RadioGridField)



from channels.generic.websockets import WebsocketConsumer
import json
from otree.models import Session
from serializer_ext.serializers import SessionSerializer
from rest_framework import generics
from django.urls import reverse
from serializer_ext.views import get_full_file_name, get_file_name, get_random_code


class JsonLoader(WebsocketConsumer):
    url_pattern = (
        r'^/jsonloader/$')
        # '/session/(?P<session_code>[a-zA-Z0-9_-]+)' +
        # '$')

    def clean_kwargs(self, kwargs):
        self.session = self.kwargs['session_code']

    def receive(self, text=None, bytes=None, **kwargs):
        json_data = json.loads(text)
        session_code =json_data.get('session_code')
        if session_code:
            session = Session.objects.filter(code=session_code)
            result = SessionSerializer(session, many=True, )
            random_code = get_random_code()
            filename = get_file_name(session_code, random_code)
            with open(get_full_file_name(filename), "w+") as f:
                f.write(json.dumps(result.data, indent=4))
                download_link = reverse('download_json', kwargs={'session_code': session_code,
                                                                 'random_code': random_code})
            self.send({'file_ready': True,
                       'download_link': download_link,
                       'session_code': session_code})


    def connect(self, message, **kwargs):
        ...
        # print('Someone connected')


    def send(self, content):
        self.message.reply_channel.send({'text': json.dumps(content)})



from rest_framework import serializers
from otree.models.subsession import BaseSubsession
from otree.models.group import BaseGroup
from otree.models.player import BasePlayer
from otree.models import Session, Participant

# we filter out these 'technical' fields.
block_fields = ['_gbat_arrived', '_gbat_grouped', '_index_in_subsessions', '_index_in_pages', '_index_in_pages',
                '_waiting_for_ids', '_last_page_timestamp', '_last_request_timestamp', 'is_on_wait_page',
                '_current_page_name', '_current_app_name', '_round_number', '_current_form_page_url', '_max_page_index',
                '_browser_bot_finished', '_pre_create_id', '_admin_report_app_names', '_admin_report_num_rounds',
                '_anonymous_code']


class oTreeSerializer(serializers.ModelSerializer):
    class Meta:
        fields = ['pk', ]

    def build_serializer(self, class_name, related_model, base_serializer):
        class NestedSerializer(oTreeSerializer):
            class Meta:
                model = related_model
                fields = '__all__'

        field_class = type(class_name, (NestedSerializer, base_serializer), {})
        return field_class

    def get_field_names(self, declared_fields, info):
        # we filter out all related models not to be in an everlasting loop. We add some of them back in
        # get_deeper_models later
        names = [f.name for f in self.Meta.model._meta.get_fields() if not f.related_model]
        res = list(super().get_field_names(declared_fields, info))
        res += names
        res = [r for r in res if r not in block_fields]
        return res

    def get_deeper_models(self):
        return

    def __init__(self, *args, **kwargs):
        # I am not sure it's an optimal solution but looks like a compact one right now
        model = kwargs.pop('model', None)
        if model:
            self.Meta.model = model
        super().__init__(*args, **kwargs)

    def get_fields(self):
        fields = super().get_fields()
        deeper = self.get_deeper_models()
        # here we look if there are some extra models to add. If yes, we are looping through all fields available
        # in the model with the fitting parent type. the 'field_name' is needed only for Subssession calling for
        # Group model (because there the related name has a naming convention not 'APP_MODEL' but simply 'group_set'
        if deeper:
            fff = self.Meta.model._meta.get_fields()
            for f in fff:
                if f.related_model:
                    if f.related_model.__base__ is deeper['base']:
                        if deeper.get('field_name'):
                            fname = deeper['field_name']
                        else:
                            fname = f.name
                        fields[fname] = self.build_serializer(class_name='serializer_{}'.format(fname),
                                                              related_model=f.related_model,
                                                              base_serializer=deeper['serializer'])(many=True)
        return fields


class PlayerSerializer(oTreeSerializer):
    class Meta:
        fields = ['pk', 'group', 'subsession']


class GroupSerializer(oTreeSerializer):
    ...


class ParticipantSerializer(oTreeSerializer):
    def get_deeper_models(self):
        return {
            'base': BasePlayer,
            'serializer': PlayerSerializer

        }

    class Meta:
        model = Participant


class SubSessionSerializer(oTreeSerializer):
    class Meta:
        fields = ()

    def get_deeper_models(self):
        return {
            'base': BaseGroup,
            'serializer': GroupSerializer,
            'field_name': 'group_set',
        }


class SessionSerializer(oTreeSerializer):
    participant_set = ParticipantSerializer(many=True)

    class Meta:
        model = Session
        fields = ('id', 'is_demo', 'num_participants', 'code', 'vars',
                  'participant_set',
                  )

    def get_deeper_models(self):
        return {
            'base': BaseSubsession,
            'serializer': SubSessionSerializer

        }



import re
from django.forms.extras.widgets import SelectDateWidget
from django.forms.widgets import Widget, Select, MultiWidget
from django.utils.safestring import mark_safe

__all__ = ('SelectTimeWidget', 'SplitSelectDateTimeWidget')

# Attempt to match many time formats:
# Example: "12:34:56 P.M."  matches:
# ('12', '34', ':56', '56', 'P.M.', 'P', '.', 'M', '.')
# ('12', '34', ':56', '56', 'P.M.')
# Note that the colon ":" before seconds is optional, but only if seconds are omitted
time_pattern = r'(\d\d?):(\d\d)(:(\d\d))? *([aApP]\.?[mM]\.?)?$'

RE_TIME = re.compile(time_pattern)
# The following are just more readable ways to access re.matched groups:
HOURS = 0
MINUTES = 1
SECONDS = 3
MERIDIEM = 4


class SelectTimeWidget(Widget):
    """
    A Widget that splits time input into <select> elements.
    Allows form to show as 24hr: <hour>:<minute>:<second>, (default)
    or as 12hr: <hour>:<minute>:<second> <am|pm>

    Also allows user-defined increments for minutes/seconds
    """
    hour_field = '%s_hour'
    minute_field = '%s_minute'
    second_field = '%s_second'
    meridiem_field = '%s_meridiem'
    twelve_hr = False  # Default to 24hr.

    def __init__(self, attrs=None, hour_step=None, minute_step=None, second_step=None, twelve_hr=False):
        """
        hour_step, minute_step, second_step are optional step values for
        for the range of values for the associated select element
        twelve_hr: If True, forces the output to be in 12-hr format (rather than 24-hr)
        """
        self.attrs = attrs or {}

        if twelve_hr:
            self.twelve_hr = True  # Do 12hr (rather than 24hr)
            self.meridiem_val = 'a.m.'  # Default to Morning (A.M.)

        if hour_step and twelve_hr:
            self.hours = range(1, 13, hour_step)
        elif hour_step:  # 24hr, with stepping.
            self.hours = range(0, 24, hour_step)
        elif twelve_hr:  # 12hr, no stepping
            self.hours = range(1, 13)
        else:  # 24hr, no stepping
            self.hours = range(0, 24)

        if minute_step:
            self.minutes = range(0, 60, minute_step)
        else:
            self.minutes = range(0, 60)

        if second_step:
            self.seconds = range(0, 60, second_step)
        else:
            self.seconds = range(0, 60)

    def render(self, name, value, attrs=None):
        try:  # try to get time values from a datetime.time object (value)
            hour_val, minute_val, second_val = value.hour, value.minute, value.second
            if self.twelve_hr:
                if hour_val >= 12:
                    self.meridiem_val = 'p.m.'
                else:
                    self.meridiem_val = 'a.m.'
        except AttributeError:
            hour_val = minute_val = second_val = 0
            if isinstance(value, str):
                match = RE_TIME.match(value)
                if match:
                    time_groups = match.groups();
                    hour_val = int(time_groups[HOURS]) % 24  # force to range(0-24)
                    minute_val = int(time_groups[MINUTES])
                    if time_groups[SECONDS] is None:
                        second_val = 0
                    else:
                        second_val = int(time_groups[SECONDS])

                    # check to see if meridiem was passed in
                    if time_groups[MERIDIEM] is not None:
                        self.meridiem_val = time_groups[MERIDIEM]
                    else:  # otherwise, set the meridiem based on the time
                        if self.twelve_hr:
                            if hour_val >= 12:
                                self.meridiem_val = 'p.m.'
                            else:
                                self.meridiem_val = 'a.m.'
                        else:
                            self.meridiem_val = None

        # If we're doing a 12-hr clock, there will be a meridiem value, so make sure the
        # hours get printed correctly
        if self.twelve_hr and self.meridiem_val:
            if self.meridiem_val.lower().startswith('p') and hour_val > 12 and hour_val < 24:
                hour_val = hour_val % 12
        elif hour_val == 0:
            hour_val = 12

        output = []
        if 'id' in self.attrs:
            id_ = self.attrs['id']
        else:
            id_ = 'id_%s' % name

        # For times to get displayed correctly, the values MUST be converted to unicode
        # When Select builds a list of options, it checks against Unicode values
        hour_val = u"%.2d" % hour_val
        minute_val = u"%.2d" % minute_val
        second_val = u"%.2d" % second_val

        hour_choices = [("%.2d" % i, "%.2d" % i) for i in self.hours]
        local_attrs = self.build_attrs(id=self.hour_field % id_)
        select_html = Select(choices=hour_choices).render(self.hour_field % name, hour_val, local_attrs)
        output.append(select_html)

        minute_choices = [("%.2d" % i, "%.2d" % i) for i in self.minutes]
        local_attrs['id'] = self.minute_field % id_
        select_html = Select(choices=minute_choices).render(self.minute_field % name, minute_val, local_attrs)
        output.append(select_html)

        second_choices = [("%.2d" % i, "%.2d" % i) for i in self.seconds]
        local_attrs['id'] = self.second_field % id_
        select_html = Select(choices=second_choices).render(self.second_field % name, second_val, local_attrs)
        output.append(select_html)

        if self.twelve_hr:
            #  If we were given an initial value, make sure the correct meridiem gets selected.
            if self.meridiem_val is not None and self.meridiem_val.startswith('p'):
                meridiem_choices = [('p.m.', 'p.m.'), ('a.m.', 'a.m.')]
            else:
                meridiem_choices = [('a.m.', 'a.m.'), ('p.m.', 'p.m.')]

            local_attrs['id'] = local_attrs['id'] = self.meridiem_field % id_
            select_html = Select(choices=meridiem_choices).render(self.meridiem_field % name, self.meridiem_val,
                                                                  local_attrs)
            output.append(select_html)

        return mark_safe(u'\n'.join(output))

    def id_for_label(self, id_):
        return '%s_hour' % id_

    id_for_label = classmethod(id_for_label)

    def value_from_datadict(self, data, files, name):
        # if there's not h:m:s data, assume zero:
        h = data.get(self.hour_field % name, 0)  # hour
        m = data.get(self.minute_field % name, 0)  # minute
        s = data.get(self.second_field % name, 0)  # second

        meridiem = data.get(self.meridiem_field % name, None)

        # NOTE: if meridiem is None, assume 24-hr
        if meridiem is not None:
            if meridiem.lower().startswith('p') and int(h) != 12:
                h = (int(h) + 12) % 24
            elif meridiem.lower().startswith('a') and int(h) == 12:
                h = 0

        if (int(h) == 0 or h) and m and s:
            return '%s:%s:%s' % (h, m, s)

        return data.get(name, None)


class SplitSelectDateTimeWidget(MultiWidget):
    """
    MultiWidget = A widget that is composed of multiple widgets.

    This class combines SelectTimeWidget and SelectDateWidget so we have something
    like SpliteDateTimeWidget (in django.forms.widgets), but with Select elements.
    """

    def __init__(self, attrs=None, hour_step=None, minute_step=None, second_step=None, twelve_hr=None, years=None):
        """ pass all these parameters to their respective widget constructors..."""
        widgets = (SelectDateWidget(attrs=attrs, years=years),
                   SelectTimeWidget(attrs=attrs, hour_step=hour_step, minute_step=minute_step, second_step=second_step,
                                    twelve_hr=twelve_hr))
        super(SplitSelectDateTimeWidget, self).__init__(widgets, attrs)

    def decompress(self, value):
        if value:
            return [value.date(), value.time().replace(microsecond=0)]
        return [None, None]

    def format_output(self, rendered_widgets):
        """
        Given a list of rendered widgets (as strings), it inserts an HTML
        linebreak between them.

        Returns a Unicode string representing the HTML for the whole lot.
        """
        rendered_widgets.insert(-1, '<br/>')
        return u''.join(rendered_widgets)


from django.conf.urls import url, include

from otree_tools.views import *
from serializer_ext.views import DownloadJson

views_to_add = [
    # Trackers views:
    FocusEventList,
    FocusPerPageReport,
    EnterExitEventList,
    TempFileCSVExport,

    # Export views
    ListPVarsView,
    PVarsCSVExport,
    SpecificSessionDataView,
    DownloadJson,
    AllSessionsList,

]
urlpatterns = [url(i.url_pattern, i.as_view(), name=i.url_name) for i in views_to_add]



# -*- coding: utf-8 -*-

import re

from django import VERSION
from django.utils import six
from django.utils.safestring import mark_safe


def add_meta_class(field):
    if VERSION < (1, 8):
        from django.db.models import SubfieldBase
        return six.add_metaclass(SubfieldBase)(field)
    return field


def widget_render(rendered):
    if VERSION < (1, 8):
        rendered = rendered.replace('li', 'td')
        rendered = re.compile(r'<ul.+>|</ul>').sub('', rendered)
        return mark_safe(rendered)
    return rendered


def get_val_from_obj(self, obj):
    if VERSION < (2, 0):
        return self._get_val_from_obj(obj)
    if obj is not None:
        return getattr(obj, self.attname)
    else:
        return self.get_default()



from otree.models_concrete import PageCompletion
from otree_tools.models import Enter, Exit, FocusEvent, focus_enter_codes, focus_exit_codes
from datetime import timedelta
import logging
from django.db.models import F, ExpressionWrapper, DurationField, Sum
from otree_tools import cp

logger = logging.getLogger('otree_tools.time.utils')


def get_seconds_per_page(player, page_name, ):
    participant = player.participant
    subsession_pk = player.subsession.pk
    try:
        page_info = PageCompletion.objects.get(participant=participant, subsession_pk=subsession_pk,
                                               page_name=page_name)
    except PageCompletion.DoesNotExist:
        raise ValueError('Cannot find info about this page')
    return page_info.seconds_on_page


def get_time_per_page(player, page_name):
    """Returns time per page as measured by tracking_time tag on a corresponding page."""

    tot_exits = Exit.objects.filter(player_id=player.pk,
                                    participant=player.participant,
                                    page_name=page_name,
                                    enter__isnull=False,
                                    ).aggregate(diff=Sum(ExpressionWrapper(F('timestamp') - F('enter__timestamp'),
                                                                           output_field=DurationField())))['diff']

    return tot_exits or timedelta()



def get_focused_time(player, page_name):
    return FocusEvent.objects.get_focused_time_per_page(player.id, player.participant.id, page_name)


def get_unfocused_time(player, page_name):
    return FocusEvent.objects.get_unfocused_time_per_page(player.id, player.participant.id, page_name)



from django.forms import widgets
import numbers
import math
from otree.widgets import CheckboxSelectMultiple

# for decompression
divider = ':'


class NoCheckboxCheckbox(CheckboxSelectMultiple):
    input_type = 'radiobutton'
    template_name = 'otree_tools/widgets/multiple_choice_select.html'
    option_template_name = 'otree_tools/widgets/multiple_choice_option.html'

    class Media:
        css = {'all': ('css/otree_tools_widgets.css',)}

    def build_attrs(self, base_attrs, extra_attrs=None):
        b = super().build_attrs(base_attrs, extra_attrs)
        b['class'] = 'form-check-input otree-tools-checkbox'
        return b


class OtherTextInput(widgets.TextInput):
    is_required = False

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        if context['widget']['attrs'].get('required'):
            context['widget']['attrs'].pop('required')
        context['other_text_input'] = name
        return context


class OtherSelectorWidget(widgets.MultiWidget):
    template_name = 'otree_tools/widgets/multiwidget.html'

    def __init__(self, other_val='other', choices=None, attrs=None):
        self.choices = choices
        self.other_val = other_val;
        _widgets = (
            widgets.RadioSelect(choices=choices, attrs={'class': 'other_wg_radio_group'}),
            OtherTextInput(attrs={'class': 'hidden_text_group', }, ),
        )
        super().__init__(_widgets, attrs)

    def get_context(self, name, value, attrs):
        con = super().get_context(name, value, attrs)
        con['wrap_label'] = True
        con['other_val'] = self.other_val
        # this one is needed only for a yet empty field
        if isinstance(value, list):
            con['widget']['show_other_inbox'] = value[0] == self.other_val
        else:
            if isinstance(value, str):
                con['widget']['show_other_inbox'] = self.decompress(value)[0] == self.other_val
        return con

    def decompress(self, value):
        # TODO what if someone starts his other answer with trailing divider? ignore for now

        if value:
            split_value = value.split(divider)
            if len(split_value) > 1:
                is_other = split_value[0] == self.other_val
                the_rest = ''.join(split_value[1:])
            else:
                is_other = False
                the_rest = ''
            if is_other:
                return [self.other_val, the_rest]
            else:
                return [value, '']
        return [None, None, ]

    def format_output(self, rendered_widgets):
        return ''.join(rendered_widgets)

    def value_from_datadict(self, data, files, name):
        radio_data = self.widgets[0].value_from_datadict(data, files, name + '_0')
        text_data = self.widgets[1].value_from_datadict(data, files, name + '_1')
        return [radio_data, text_data]


permitted_info_block = ['left', 'right']
non_permitted_block_msg = 'Choose between two options: {}'.format(' or '.join(permitted_info_block))
no_range_set_err_msg = 'Both max and min parameters should be set to use this slider'


class AdvancedSliderWidget(widgets.NumberInput):
    template_name = 'otree_tools/widgets/tickslider.html'
    # TODO: provide a chance for user not to give min and max
    # default_min = 0
    # default_max = 10
    # default_range = 10
    default_nsteps = 10
    default_med_value = 5
    ndigits = 0
    suffix = ''

    def __init__(self, show_ticks=True, show_value=True, show_block='left', ndigits=None, *args, **kwargs):
        self.show_ticks = show_ticks
        self.show_value = show_value
        self.ndigits = ndigits
        assert show_block in permitted_info_block, non_permitted_block_msg
        self.show_block = show_block
        super().__init__(*args, **kwargs)

    def set_steps(self, smin, smax):
        slider_range = smax - smin
        step = round(slider_range / self.default_nsteps, 2)
        return step

    def set_digits(self, step):
        if self.ndigits is None:
            if isinstance(step, int):
                self.ndigits = 0
                return
            frac, _ = math.modf(step)
            strfrac = str(frac).split('.')[1]
            self.ndigits = len(strfrac)

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        a_ = context['widget']['attrs']
        assert all(i in a_ for i in ['min', 'max']), no_range_set_err_msg
        # TODO: allow user to skip max/min providing default options (partly based on step/tick_interval options)
        # wi_attrs.setdefault('min', self.default_min)
        # wi_attrs.setdefault('max', self.default_max)
        if not isinstance(a_.get('step'), numbers.Number):
            a_['step'] = self.set_steps(a_['min'], a_['max'])
        self.set_digits(a_['step'])
        a_['ndigits'] = self.ndigits

        a_.setdefault('tick_interval', a_['step'])
        a_.setdefault('secondary_ticks', True)
        a_.setdefault('show_ticks', True)
        self.default_med_value = round((a_['min'] + a_['max']) / 2, self.ndigits)
        if self.ndigits == 0:
            self.default_med_value = int(self.default_med_value)
        a_['slider_start_value'] = value if value is not None else self.default_med_value
        a_.setdefault('suffix', self.suffix)
        a_['show_value'] = self.show_value
        a_['show_block'] = self.show_block
        return context



        # TODO: when Chris includes form.media to the template, stop adding jquery-ui to each field which is absu
        # class Media:
        #     css = {
        #         'all': ('jquery-ui/jquery-ui.min.css',
        #                 'css/slider.css',)
        #     }
        #     js = ('jquery-ui/jquery-ui.min.js',
        #           'js/slider.js',)



import json
from datetime import datetime
from otree_tools.models import (Enter, Exit, FocusEvent, focus_enter_codes,
                                focus_exit_codes, allowed_export_tracker_requests)
from channels.generic.websockets import JsonWebsocketConsumer
from django.contrib.contenttypes.models import ContentType
from otree.models import Participant
from otree.models_concrete import ParticipantToPlayerLookup
import logging
from otree_tools import cp
from otree_tools.prepare_export_data import FileMaker
from django.db.utils import IntegrityError
from django.template.loader import render_to_string

logger = logging.getLogger('otree_tools.consumers')

"""
Export of timing objects is relatively slow thing.
We should check whether redis is available - if yes, we can try to use hyeu to create a task that will call
initiate downloading as soon as the file is ready.

An alternative solution would be to send a signal to database, and start preparing file. 
as soon as it is ready the signal is sent back, which allows to download it.
A
"""


class GeneralTracker(JsonWebsocketConsumer):
    tracker_url_kwargs = r'(?P<participant_code>[a-zA-Z0-9_-]+)/(?P<page_name>[a-zA-Z0-9_-]+)/(?P<page_index>\d+)$'

    def clean_kwargs(self):
        self.participant_code = self.kwargs['participant_code']
        self.page_name = self.kwargs['page_name']
        self.page_index = self.kwargs['page_index']

    def get_participant(self):
        self.clean_kwargs()
        try:
            return Participant.objects.get(code__exact=self.participant_code)
        except Participant.DoesNotExist:
            return

    def get_player_and_app(self):
        empty_ret = (None, None, None)
        participant = self.get_participant()
        if participant:
            try:
                participant_lookup_item = participant.participanttoplayerlookup_set.get(
                    page_index=self.page_index)
                app_name = participant_lookup_item.app_name
                player_pk = participant_lookup_item.player_pk

                player_content_type = ContentType.objects.get(app_label=app_name, model='player')
                player = player_content_type.get_object_for_this_type(pk=player_pk)
                return participant, app_name, player

            except ParticipantToPlayerLookup.DoesNotExist:
                logger.error('Cannot find participant for this page!')
        return empty_ret


class TimeTracker(GeneralTracker):
    # TODO: deal with page unloaded registered BEFORE form submitted
    url_pattern = (r'^/timetracker/' + GeneralTracker.tracker_url_kwargs)

    def receive(self, content, **kwargs):
        raw_content = json.loads(content)
        raw_time = raw_content['timestamp']
        timestamp = datetime.fromtimestamp(raw_time / 1000)
        event_type = raw_content['eventtype']
        wait_for_images = raw_content.get('wait_for_images', True)
        participant, app_name, player = self.get_player_and_app()
        filter_params = {'page_name': self.page_name,
                         'participant': participant,
                         'app_name': app_name, }
        general_params = {**filter_params,
                          'player': player,
                          'timestamp': timestamp,
                          'round_number': player.round_number}
        if event_type == 'enter':
            if participant is not None:
                Enter.objects.create(**general_params,
                                     wait_for_images=wait_for_images)

        if event_type == 'exit':
            if participant is not None:
                exit_type = int(raw_content['exittype'])
                try:
                    enter = Enter.objects.filter(**filter_params,
                                                 timestamp__lte=timestamp,
                                                 player_id=player.pk,
                                                 exit__isnull=True).latest()
                except Enter.DoesNotExist:
                    enter = None

                try:
                    e = Exit(**general_params,
                             exit_type=exit_type,
                             enter=enter
                             )
                    e.save()
                except IntegrityError:
                    enter = None
                    e = Exit(**general_params,
                             exit_type=exit_type,
                             enter=enter
                             )
                    e.save()

    def connect(self, message, **kwargs):
        print('Client connected to time tracker...')

    def disconnect(self, message, **kwargs):
        print('Client disconnected from time tracker...')


#

class FocusTracker(GeneralTracker):
    url_pattern = (r'^/focustracker/' + GeneralTracker.tracker_url_kwargs)

    def get_entry(self):
        filter_params = {**self.filter_params,
                         'timestamp__lte': self.timestamp,
                         'player_id': self.player.pk,
                         'closure__isnull': True
                         }
        if self.event_num_type in focus_enter_codes:
            filter_params['event_num_type__in'] = focus_exit_codes

        if self.event_num_type in focus_exit_codes:
            filter_params['event_num_type__in'] = focus_enter_codes
        entry = FocusEvent.objects.filter(**filter_params, )
        if entry.exists():
            entry = entry.latest()
        else:
            entry = None
        return entry

    def receive(self, content, **kwargs):
        raw_content = json.loads(content)
        raw_time = raw_content['timestamp']
        self.timestamp = datetime.fromtimestamp(raw_time / 1000)
        self.event_num_type = raw_content['event_num_type']
        event_desc_type = raw_content['event_desc_type']
        participant, app_name, self.player = self.get_player_and_app()
        self.filter_params = {'page_name': self.page_name,
                              'participant': participant,
                              'app_name': app_name, }
        if participant is not None:
            entry = self.get_entry()
            creating_params = {**self.filter_params,
                               'timestamp': self.timestamp,
                               'player': self.player,
                               'round_number': self.player.round_number,
                               'event_desc_type': event_desc_type,
                               "event_num_type": self.event_num_type,
                               "entry": entry}
            try:
                FocusEvent.objects.create(**creating_params)
            except IntegrityError:
                creating_params['entry'] = None
                FocusEvent.objects.create(**creating_params)


def connect(self, message, **kwargs):
    print('Client connected to focus tracker...')


class ExportTracker(JsonWebsocketConsumer):
    url_pattern = r'^/exporttracker/(?P<inner_group_id>.*)$'
    progress_block_name = 'tester/includes/progress_block.html'

    def connection_groups(self, **kwargs):
        return [self.kwargs['inner_group_id']]

    def receive(self, content, **kwargs):
        raw_content = json.loads(content)
        request = raw_content['request']
        tracker_type = raw_content['tracker_type']

        if request == 'export':
            msg = {'button': render_to_string(self.progress_block_name, {})}
            self.send(msg)
            file_maker = FileMaker(channel_to_response=self.kwargs['inner_group_id'],
                                   tracker_type=tracker_type)
            file_maker.get_data()

    def connect(self, message, **kwargs):
        logger.info('Client connected to export tracker...')

    def disconnect(self, message, **kwargs):
        logger.info('Client disconnected from export tracker...')
#



""" things to do in mturk module
 when a person has a session run in mturk, he should be able to manage the hit from in there:
  - sending extra bonuses;
  - manage qualifications
  - reacept rejected assignments
  - send messages

"""


from django import template
from django.template import TemplateSyntaxError, Context
from otree_tools.models.models import FOCUS_EVENT_TYPES


class ButtonTagError(Exception):
    pass


register = template.Library()
# two following random strings are needed to prevent usage of tracking_time and tracking_focus twice
# see more here: https://stackoverflow.com/questions/51786795/check-that-tag-is-used-in-template-only-once/
tracking_time_code, tracking_focus_code = 'xBl4fTBP8hAsg61o1gJa', 'Q4YviMPgcVUjbbm05C6s'


def universal_tracker(context, tagname):
    context['page_index'] = context['view']._index_in_pages
    context['page_name'] = context['view'].__class__.__name__
    code = globals()[f'{tagname}_code']
    if code in context:
        formatted_tag_name = '{{% {} %}}'.format(tagname)
        raise Exception(f'{formatted_tag_name} is already used')
    context[code] = True
    return context


@register.inclusion_tag('otree_tools/tags/TimeTracker.html', takes_context=True)
def tracking_time(context, wait_for_images=True, *args, **kwargs):
    context['wait_for_images'] = wait_for_images
    return universal_tracker(context, 'tracking_time')


@register.inclusion_tag('otree_tools/tags/FocusTracker.html', takes_context=True, name='tracking_focus')
def tracking_focus_func(context, *args, **kwargs):
    c = universal_tracker(context, 'tracking_focus')
    c['type_correspondence'] = dict(FOCUS_EVENT_TYPES)
    return c


@register.inclusion_tag('otree_tools/tags/Button.html', takes_context=True)
def button(context, label='', *args, **kwargs):
    for kwarg in kwargs:
        raise ButtonTagError(
            # need double {{ to escape because of .format()
            '{{% button %}} tag received unrecognized parameter "{}"'.format(kwarg)
        )
    context['label'] = label
    return context


@register.inclusion_tag('otree_tools/tags/ConfirmButton.html', takes_context=True, name='confirm_button')
def confirm_button(context,
                   title='Info',
                   message='Are you sure you would like to proceed?',
                   main_button='Next',
                   yes_button='Yes',
                   no_button='No'):
    c = context
    c['title'] = title
    c['message'] = message
    c['main_button'] = main_button
    c['yes_button'] = yes_button
    c['no_button'] = no_button
    return c



# -*- coding: utf-8 -*-

from django import VERSION

if VERSION < (1, 11):
    from django.forms.widgets import RadioSelect, MultiWidget, RadioFieldRenderer
    from django.utils.safestring import mark_safe
    from django.template.loader import render_to_string

    from otree_tools.radiogrid.compat import widget_render


    class RadioChoiceFieldRenderer(RadioFieldRenderer):
        outer_html = '{content}'
        inner_html = '<td>{choice_value}{sub_widgets}</td>'

        def __init__(self, name, value, attrs, choices):
            attrs['class'] = ''
            super(RadioChoiceFieldRenderer, self).__init__(name, value, attrs, choices)

        def render(self):
            rendered = super(RadioChoiceFieldRenderer, self).render()
            return widget_render(rendered)


    class RadioRadioSelect(RadioSelect):
        renderer = RadioChoiceFieldRenderer


    class RadioGridWidget(MultiWidget):
        input_type = 'grid'

        def __init__(self, rows, values, attrs=None):
            self.rows = rows
            self.values = values
            choices = [(k, '') for k, _ in values]
            widgets = [RadioRadioSelect(choices=choices, attrs=attrs) for _ in rows]
            super(RadioGridWidget, self).__init__(widgets, attrs)

        def decompress(self, value):
            if value:
                return value.split(',')
            return [None for _ in self.rows]

        def format_output(self, rendered_widgets):
            widgets = {}
            for i, v in enumerate(rendered_widgets):
                widgets[i] = v
            return mark_safe(render_to_string('radiogrid/radiogrid_widget.html', {
                'rows': [(v[1], widgets[i]) for i, v in enumerate(self.rows)],
                'values': self.values,
                'attrs': self.attrs
            }))
else:
    from django.forms import RadioSelect, MultiWidget
    from django.utils.safestring import mark_safe
    from django.template.loader import render_to_string


    class RadioRadioSelect(RadioSelect):
        template_name = 'radiogrid/radiogrid_input.html'


    class RadioGridWidget(MultiWidget):
        input_type = 'grid'
        template_name = 'radiogrid/radiogrid_widget.html'

        def __init__(self, rows, values, attrs=None):
            self.rows = rows
            self.values = values
            choices = [(k, '') for k, _ in values]
            widgets = [RadioRadioSelect(choices=choices, attrs=attrs) for _ in rows]
            super(RadioGridWidget, self).__init__(widgets, attrs)

        def get_context(self, name, value, attrs):
            context = super(RadioGridWidget, self).get_context(name, value, attrs)
            widgets = context['widget']['subwidgets']
            return {
                'rows': [(v[1], widgets[i]) for i, v in enumerate(self.rows)],
                'values': self.values,
                'attrs': self.attrs
            }

        def decompress(self, value):
            if value:
                return value.split(',')
            return [None for _ in self.rows]



from otree.models import Session
from django.views.generic import TemplateView, View
from django.shortcuts import render
from .serializers import SessionSerializer
from rest_framework import generics
from rest_framework.renderers import JSONRenderer, BrowsableAPIRenderer
import random
import string
import os
from django.http import HttpResponse
from wsgiref.util import FileWrapper


def get_full_file_name(filename):
    return 'serializer_ext/temp/{}'.format(filename)


def get_random_code():
    return ''.join(random.choice(string.ascii_letters) for i in range(10))


def get_file_name(session_code, random_code):
    return 'session_data_{}_{}.json'.format(session_code, random_code)


def downloadable_filename(session_code):
    return 'session_data_{}.json'.format(session_code)


# TODO: move up imports
import json
from django.template import Context, loader


class DownloadJson(TemplateView):
    template_name = 'download_json.json'
    url_pattern = r'^download_json/(?P<session_code>\w+)$'
    url_name = 'download_json'

    def get(self, request, *args, **kwargs):
        response = HttpResponse(content_type='application/json')
        session_code = self.kwargs['session_code']
        filename = '{}_json_data.json'.format(session_code)
        response['Content-Disposition'] = 'attachment; filename="{}"'.format(filename)
        session = Session.objects.filter(code=session_code)
        result = SessionSerializer(session, many=True, )
        data = json.dumps(result.data, indent=4)
        t = loader.get_template(self.template_name)
        c = {
            'data': data,
        }
        response.write(t.render(c))
        return response


# the view to get a list of all sessions
class AllSessionsList(TemplateView):
    template_name = 'serializer_ext/all_session_list.html'
    url_name = 'json_sessions_list'
    url_pattern = r'^sessions_list_for_json/$'
    display_name = 'Exporting data to JSON'

    def get(self, request, *args, **kwargs):
        all_sessions = Session.objects.all()
        return render(request, self.template_name, {'sessions': all_sessions})



# original code is by Anton Shurashov (github user Sinkler)
# https://github.com/Sinkler/django-radiogrid

from .widgets import RadioGridWidget
from .fields import RadioGridFormField
from .db import RadioGridField

__version__ = '1.0.4'



from channels.routing import route_class
from .consumers import (TimeTracker, FocusTracker, ExportTracker)

trackers = [TimeTracker, FocusTracker, ExportTracker]
channel_routing = [route_class(i, path=i.url_pattern, ) for i in trackers]



__version__ = '1.1.1'

"""The following is needed just for naive debugging."""
from termcolor import colored
import shutil


def cp(*x, color='red', center=True):
    columns = shutil.get_terminal_size().columns
    x = ' '.join([str(i) for i in x])
    if center:
        x = x.center(columns, '-')
    print(colored(x, color, 'on_cyan', attrs=['bold']))



from django.conf.urls import url, include
from serializer_ext import views as v
from django.conf import settings
from django.contrib.auth.decorators import login_required

urlpatterns = [
    url(r'^download_json/(?P<session_code>\w+)/(?P<random_code>\w+)/$', v.DownloadJson.as_view(), name='download_json'),
]

